#lang typed/racket/base

;; ff34
;; parallel (restricted splitting), list-based, typed
;; 1st attempt at reducing blocking, using types for unboxed arithmetic
;; 2nd attempt at reducing blocking, rewrote `append' in Racket

(provide fft3/time
         fft3
         fft3/depth
         append
         decimate-in-time)

(require (only-in racket/math pi)
         (only-in racket/future future touch)
         racket/fixnum
         racket/flonum)
(require/typed "fft2.rkt" [init-d (-> Fixnum)])

(: fft3/time ((Listof Float-Complex) -> (Listof Float-Complex)))
(define (fft3/time as) (time (fft3 as)))

;; START fft3
(: fft3 ((Listof Float-Complex)
         ->
         (Listof Float-Complex)))
(define (fft3 as)
  (fft3/depth as (init-d)))

(: fft3/depth ((Listof Float-Complex)
               Integer
               -> 
               (Listof Float-Complex)))
(define (fft3/depth as d)
  (cond
    [(= (length as) 1) as]
    [(= d 0) (fft3/seq as)]
    [else
     (define-values (ss ts) 
       (decimate-in-time as))
     (define bs
       (future
        (λ ()
          (fft3/depth ss (- d 1)))))
     (define cs
       (future
        (λ ()
          (twiddle-factor
           (fft3/depth ts (- d 1))))))
     (append
      (for/list: : (Listof Float-Complex)
        ([b (in-list (touch bs))]
         [c (in-list (touch cs))])
        (+ b c))
      (for/list: : (Listof Float-Complex)
        ([b (in-list (touch bs))]
         [c (in-list (touch cs))])
        (- b c)))]))
;; STOP fft3

(: fft3/seq : (Listof Float-Complex) -> (Listof Float-Complex))
(define (fft3/seq as)
  (cond [(= (length as) 1) as]
        [else
         (define-values (ss ts) (decimate-in-time as))
         (define bs (fft3/seq ss))
         (define cs (twiddle-factor (fft3/seq ts)))
         (append (for/list: : (Listof Float-Complex)
                            ([b (in-list bs)] [c (in-list cs)])
                            (+ b c))
                 (for/list: : (Listof Float-Complex)
                            ([b (in-list bs)] [c (in-list cs)])
                            (- b c)))]))

(: decimate-in-time : (∀ (A) ((Listof A) -> (values (Listof A) (Listof A)))))
(define (decimate-in-time as)
  (for/fold: ([evens : (Listof A) '()]
              [odds  : (Listof A) '()])
             ([i (in-naturals)]
              [a (in-list (reverse as))])
    (if (odd? i)
        (values (cons a evens) odds)
        (values evens (cons a odds)))))

(: twiddle-factor : (Listof Float-Complex) -> (Listof Float-Complex))
(define (twiddle-factor cs)
  (define n (->fl (length cs)))
  (for/list ([k (in-naturals)]
             [c (in-list cs)])
    (* c (exp (/ (* pi 0.0+1.0i (->fl k)) 
                 n)))))

(: append : (All (A) ((Listof A) (Listof A) -> (Listof A))))
(define (append l1 l2)
  (cond
    [(null? l1) l2]
    [(null? l2) l1]
    [else
     (let app-loop ([l1 l1]
                    [l2 l2])
       (cond
         [(null? l1) l2]
         [else
          (cons (car l1) 
                (app-loop (cdr l1) l2))]))]))


(provide run-fft3)
(: run-fft3 : (Listof Float-Complex)
   -> (values (Listof Any) Integer Integer Integer))
(define (run-fft3 l)
  (run-fft3/d l (init-d)))

(provide run-fft3/d)
(: run-fft3/d : (Listof Float-Complex) Integer
   -> (values (Listof Any) Integer Integer Integer))
(define (run-fft3/d l d)
  (collect-garbage) (collect-garbage) (collect-garbage)
  (collect-garbage) (collect-garbage) (collect-garbage)
  ((inst time-apply (Listof Float-Complex) (Listof Float-Complex) Integer)
   fft3/depth (list l d)))

