#lang typed/racket/base

;; fft4
;; parallel (restricted splitting), list-based, typed
;; 1st attempt at reducing blocking, using types for unboxed arithmetic
;; 2nd attempt at reducing blocking, rewrote `append' in Racket
;; A bit of loop fusion, to reduce number of traversals.
;; Inlined fourier-shift to remove an intermediate list and keep complexes
;; unboxed for longer.

(provide fft4 fft4/depth fft4/time)

(require (only-in racket/math pi)
         (only-in racket/future future touch)
         racket/fixnum
         racket/flonum)

(require/typed racket/future [processor-count (-> Integer)])
(require/typed "fft2.rkt" [init-d (-> Fixnum)])
(require "fft3.rkt")

(: fft4/time : (Listof Float-Complex) -> (Listof Float-Complex))
(define (fft4/time as) (time (fft4 as)))

(: fft4 : (Listof Float-Complex) -> (Listof Float-Complex))
(define (fft4 as)
  (fft4/depth as (init-d)))

;; START fft4
(: fft4/depth ((Listof Float-Complex) 
               Integer
               -> 
               (Listof Float-Complex)))
(define (fft4/depth as d)
  (cond 
    [(= (length as) 1) as]
    [(= d 0) (fft4/seq as)]
    [else
     (define-values (ss ts)
       (decimate-in-time as))
     (define bs
       (future
        (λ () 
          (fft4/depth ss (- d 1)))))
     (define cs
       (future
        (λ ()
          (fft4/depth ts (- d 1)))))
     (define n/2 (->fl (length (touch cs))))
     (define-values (l r)
       (for/lists: 
         ([l : (Listof Float-Complex)]
          [r : (Listof Float-Complex)])
         ([b (in-list (touch bs))] 
          [c (in-list (touch cs))]
          [k (in-naturals)])
         (define twiddle-c
           (* c (exp (/ (* pi 0.0+1.0i 
                           (->fl k))
                        n/2))))
         (values (+ b twiddle-c) 
                 (- b twiddle-c))))
     (append l r)]))
;; STOP fft4

(: fft4/seq : (Listof Float-Complex) -> (Listof Float-Complex))
(define (fft4/seq as)
  (cond [(= (length as) 1) as]
        [else
         (define-values (ss ts) (decimate-in-time as))
         (define bs (fft4/seq ss))
         (define cs (fft4/seq ts))
         (define n/2 (->fl (length cs)))
         (define-values (l r)
           (for/lists: ([l : (Listof Float-Complex)]
                        [r : (Listof Float-Complex)])
                 ([b (in-list bs)] [c (in-list cs)]
                  [k (in-naturals)])
             (define shifted-c
               (* c (exp (/ (* pi 0.0+1.0i (->fl k))
                            n/2))))
             (values (+ b shifted-c) (- b shifted-c))))
         (append l r)]))

(provide run-fft4)
(: run-fft4 : (Listof Float-Complex)
   -> (values (Listof Any) Integer Integer Integer))
(define (run-fft4 l)
  (run-fft4/d l (init-d)))

(provide run-fft4/d)
(: run-fft4/d : (Listof Float-Complex) Integer
   -> (values (Listof Any) Integer Integer Integer))
(define (run-fft4/d l d)
  (collect-garbage) (collect-garbage) (collect-garbage)
  (collect-garbage) (collect-garbage) (collect-garbage)
  ((inst time-apply (Listof Float-Complex) (Listof Float-Complex) Integer)
   fft4/depth (list l d)))
