#include <stdio.h>
#include <complex.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include <sys/time.h>
#include <stdlib.h>
#include <assert.h>

static double get_milliseconds();
static void print_nums(double complex as[], int len);
static double complex make_rectangular(double x, double y);
static double complex my_cexp(double complex c);
static void get_real_parts(double complex as[], const int len, double result[]);
static void get_imag_parts(double complex as[], const int len, double result[]);
static double complex *fft(double complex *as, int d, double complex ret[]);
static void fft_depth(double as_r[], 
                       double as_i[], 
                       double xs_r[], 
                       double xs_i[], 
                       int n, 
                       int start, 
                       int d);
static void decimate_in_time( double as_r[], 
                              double as_i[], 
                              double xs_r[], 
                              double xs_i[], 
                              int n_by_2, 
                              int start);
static void twiddle_factor(double cs_r[], double cs_i[], int nby2, int start);
static void *par_bs(void *arg);
static void *par_cs(void *arg);

typedef struct par_work {
  double *xs_r;
  double *xs_i;
  double *as_r;
  double *as_i;
  int nby2;
  int start;
  int d;
} par_work;


static double get_milliseconds()
{
  struct timeval now;
  gettimeofday(&now, NULL);
  return (double)now.tv_sec * 1000.0 + (double)now.tv_usec / 1000;
}

static void print_complex(double complex x)
{
  double xi = cimag(x);
  printf("%a%s%ai\n", creal(x), xi < 0 ? "" : "+", xi);
}

static void print_nums(double complex as[], int len)
{
  int i;
  double complex a;
  for (i = 0; i < len; i++) {
    print_complex(as[i]);
  }
}

void test_get_ir_parts()
{
  double complex xs[] = { 2.0 + 4.0 * I, 3.0 + 1.0 * I, 72.0 + 42.0 * I };
  double rs[3];
  double is[3];
  get_real_parts(xs, 3, rs);
  get_imag_parts(xs, 3, is);

  assert(rs[0] == 2.0);
  assert(rs[1] == 3.0);
  assert(rs[2] == 72.0);
  assert(is[0] == 4.0);
  assert(is[1] == 1.0);
  assert(is[2] == 42.0);
}

void test_make_rect()
{
  double complex xs[] = { 2.0 + 4.0 * I, 3.0 + 1.0 * I, 72.0 + 42.0 * I };
  double complex a = xs[0];
  double complex b = xs[1];
  double complex c = xs[2];

  double ar = creal(a);
  double ai = cimag(a);
  double complex aa = make_rectangular(ar, ai);
  assert(a == aa);

  double br = creal(b);
  double bi = cimag(b);
  double complex bb = make_rectangular(br, bi);
  assert(b == bb);

  double cr = creal(c);
  double ci = cimag(c);
  double complex cc = make_rectangular(cr, ci);
  assert(c == cc);
}

void test_twiddle_factor()
{
  double csr[3] = { 1.3, 1.3, 1.3 };
  double csi[3] = { 2.0, 2.0, 2.0 };
  twiddle_factor(csr, csi, 1, 0);

  assert(csr[0] == 1.3f);
  assert(csr[1] == 1.3f);
  assert(csr[2] == 1.3f);
  assert(csi[0] == 2.0f);
  assert(csi[1] == 2.0f);
  assert(csi[2] == 2.0f);
}

void run_tests()
{
  test_get_ir_parts();
  test_make_rect();
  test_twiddle_factor();
}

/* Define 'LEN', 'D', and DATA_FILE */
#include "args.h"

double complex *get_input()
{
  FILE *file = fopen(DATA_PATH, "r");
  if (file) {
    int i = 0;
    double complex *inp = (double complex*)malloc(sizeof(double complex) * LEN);
    char line[256];
    char *part;
    double real, imag;
    while (fgets(line, sizeof(line), file) != NULL) {
      part = strtok(line, "|");
      real = atof(part);
      part = strtok(NULL, "|");
      imag = atof(part);
      inp[i] = real + imag * I;

      i++;
    }

    return inp;    
  }

  return NULL;
}

int main(int argc, const char **argv)
{
#ifdef RUN_TESTS
  run_tests();
#endif

  double complex *input = get_input();
  double complex *ret = (double complex*)malloc(sizeof(double complex) * LEN);
  fft(input, D, ret);

#ifdef PRINT_RESULT
  print_nums(ret, LEN);
#endif

  return 0;
}

double complex *fft(double complex as[], int d, double complex ret[])
{
  size_t sz = sizeof(double) * LEN;
  double *as_r = (double*)malloc(sz);
  double *as_i = (double*)malloc(sz);
  get_real_parts(as, LEN, as_r);
  get_imag_parts(as, LEN, as_i);

  double *xs_r = (double*)malloc(sz);
  double *xs_i = (double*)malloc(sz);
  memset(xs_r, 0.0, sz);
  memset(xs_i, 0.0, sz);

  double start = get_milliseconds();
  fft_depth(as_r, as_i, xs_r, xs_i, LEN, 0, d);
  double end = get_milliseconds();

  int i;
  for (i = 0; i < LEN; i++) {
    ret[i] = make_rectangular(as_r[i], as_i[i]);
  }

  free(as_r);
  free(as_i);
  free(xs_r);
  free(xs_i);

#ifdef BENCHMARK 
  printf("%f", end - start);
#endif
}

static void fft_depth(double as_r[], 
                      double as_i[], 
                      double xs_r[], 
                      double xs_i[], 
                      int n, 
                      int start, 
                      int d)
{
  if (n == 1)
    return;

  int nby2 = n / 2;
  decimate_in_time(as_r, as_i, xs_r, xs_i, nby2, start);

  if (d == 0) {
    fft_depth(xs_r, xs_i, as_r, as_i, nby2, start, 0);
    fft_depth(xs_r, xs_i, as_r, as_i, nby2, start + nby2, 0);
    twiddle_factor(xs_r, xs_i, nby2, start + nby2);
  } else {
    void *status;
    pthread_t b;
    pthread_t c;
    par_work bw = { .xs_r = xs_r, 
                    .xs_i = xs_i, 
                    .as_r = as_r, 
                    .as_i = as_i, 
                    .nby2 = nby2, 
                    .start = start, 
                    .d = d - 1 };
    par_work cw = { .xs_r = xs_r, 
                    .xs_i = xs_i, 
                    .as_r = as_r, 
                    .as_i = as_i, 
                    .nby2 = nby2, 
                    .start = start + nby2, 
                    .d = d - 1 };

    int bs = pthread_create(&b, NULL, par_bs, (void*)&bw);
    int cs = pthread_create(&c, NULL, par_cs, (void*)&cw);
    bs = pthread_join(b, &status);
    cs = pthread_join(c, &status);
  }

  int k;
  for (k = 0; k < nby2; k++) {
    int sk = start + k;
    int sk2 = sk + nby2;

    assert(sk < LEN);
    assert(sk2 < LEN);

    double br = xs_r[sk];
    double bi = xs_i[sk];
    double cr = xs_r[sk2];
    double ci = xs_i[sk2];
      
    as_r[sk2] = br - cr;
    as_i[sk2] = bi - ci;
    as_r[sk] = br + cr;
    as_i[sk] = bi + ci;
   }
}

static void *par_bs(void *arg)
{
  par_work *args = (par_work*)arg;
  fft_depth(args->xs_r, 
            args->xs_i, 
            args->as_r, 
            args->as_i, 
            args->nby2, 
            args->start, 
            args->d);
  return NULL;
}

static void *par_cs(void *arg)
{
  par_work *args = (par_work*)arg;
  fft_depth(args->xs_r, 
            args->xs_i, 
            args->as_r, 
            args->as_i, 
            args->nby2, 
            args->start, 
            args->d);

  twiddle_factor(args->xs_r, args->xs_i, args->nby2, args->start);
  return NULL;
}

static void decimate_in_time( double as_r[], 
                              double as_i[], 
                              double xs_r[], 
                              double xs_i[], 
                              int nby2, 
                              int start)
{
  int i;
  for (i = 0; i < nby2; i++) {
    int si = start + i;
    int si2 = si + i;
    int si21 = si2 + 1;
    int sin2 = si + nby2;

    assert(si < LEN);
    assert(si2 < LEN);
    assert(si21 < LEN);
    assert(sin2 < LEN);

    xs_r[si] = as_r[si2];
    xs_i[si] = as_i[si2];
    xs_r[sin2] = as_r[si21];
    xs_i[sin2] = as_i[si21];
  }
}

static void twiddle_factor(double cs_r[], double cs_i[], int nby2, int start)
{
  int k;
  for (k = 0; k < nby2; k++) {
    int kstart = start + k;
    assert(kstart < LEN);

    /* M_PI is gcc-specific */
    double complex a = make_rectangular(cs_r[kstart], cs_i[kstart]);
    double complex b = M_PI * (0.0 + 1.0 * I) * (double)k;
    double complex c = b / (double)nby2;
    double complex d = cexp(c);
    double complex res = a * d;
    
    cs_r[kstart] = creal(res);
    cs_i[kstart] = cimag(res);
  }
}

/* Using this function instead of 'cexp' will 
  yield answers that exactly match Racket's 
  (cexp will not always do this with numbers 
  with more than 17 decimal places or so).  
  Need to determine the performance difference between 
  this and cexp, however. */
static double complex my_cexp(double complex c)
{
  double r, i, cos_a, sin_a;

  r = creal(c);
  i = cimag(c);

  r = exp(r);
  cos_a = cos(i);
  sin_a = sin(i);

  return r * (cos_a + (sin_a * I));
}


double complex make_rectangular(double x, double y)
{
  return x + (y * (0 + (1 * I)));
}


void get_real_parts(double complex as[], const int len, double result[])
{
  int i;
  for (i = 0; i < len; i++) {
    result[i] = creal(as[i]); 
  }
}

void get_imag_parts(double complex as[], const int len, double result[])
{
  int i;
  for (i = 0; i < len; i++) {
    result[i] = cimag(as[i]);
  }
}
